"""
FlightGear Flight Dynamics Model Network interface, version 25

See https://github.com/FlightGear/flightgear/blob/4049acd84e6ce391184716d1020c456d389f919e/src/Network/net_fdm.hxx
"""

from construct import Array, Enum, Const, Padding, Int32ub, Int32sb, Float64b, Float32b

FG_MAX_ENGINES = 4  #: Constant value from enum
FG_MAX_WHEELS = 3  #: Constant value from enum
FG_MAX_TANKS = 4  #: Constant value from enum

# Big endian
#: FDM v25 structure
fdm_struct = {
    'version': Const(25, Int32ub),
    '_padding': Padding(4),
    'lon_rad': Float64b,
    'lat_rad': Float64b,
    'alt_m': Float64b,
    'agl_m': Float32b,
    'phi_rad': Float32b,
    'theta_rad': Float32b,
    'psi_rad': Float32b,
    'alpha_rad': Float32b,
    'beta_rad': Float32b,
    'phidot_rad_per_s': Float32b,
    'thetadot_rad_per_s': Float32b,
    'psidot_rad_per_s': Float32b,
    'vcas': Float32b,  # calibrated airspeed
    'climb_rate_ft_per_s': Float32b,
    'v_north_ft_per_s': Float32b,
    'v_east_ft_per_s': Float32b,
    'v_down_ft_per_s': Float32b,
    'v_body_u': Float32b,  # ECEF velocity in body frame
    'v_body_v': Float32b,  # ECEF velocity in body frame
    'v_body_w': Float32b,  # ECEF velocity in body frame
    'A_X_pilot_ft_per_s_per_s': Float32b,
    'A_Y_pilot_ft_per_s_per_s': Float32b,
    'A_Z_pilot_ft_per_s_per_s': Float32b,
    'stall_warning': Float32b,  # 0.0 - 1.0 indicating the amount of stall
    'slip_deg': Float32b,  # slip ball deflection
    'num_engines': Int32ub,
    'eng_state': Array(FG_MAX_ENGINES,
                       Enum(Int32ub, off=0, cranking=1, running=2)
                       ),  # Engine state
    'rpm': Array(FG_MAX_ENGINES, Float32b),
    'fuel_flow_gal_per_hr': Array(FG_MAX_ENGINES, Float32b),
    'fuel_px_psi': Array(FG_MAX_ENGINES, Float32b),
    'egt_deg_F': Array(FG_MAX_ENGINES, Float32b),
    'cht_deg_F': Array(FG_MAX_ENGINES, Float32b),
    'mp_osi': Array(FG_MAX_ENGINES, Float32b),  # Manifold pressure
    'tit': Array(FG_MAX_ENGINES, Float32b),  # Turbine Inlet Temperature
    'oil_temp_deg_F': Array(FG_MAX_ENGINES, Float32b),
    'oil_px_psi': Array(FG_MAX_ENGINES, Float32b),
    'num_tanks': Int32ub,  # Max number of fuel tanks
    'fuel_quantity': Array(FG_MAX_TANKS, Float32b),
    # v24->v25 changes
    'tank_selected': Array(FG_MAX_TANKS, Int32ub),
    'capacity_m3': Array(FG_MAX_TANKS, Float64b),
    'unusable_m3': Array(FG_MAX_TANKS, Float64b),
    'density_kgpm3': Array(FG_MAX_TANKS, Float64b),
    'level_m3': Array(FG_MAX_TANKS, Float64b),
    # end changes
    'num_wheels': Int32ub,
    'wow': Array(FG_MAX_WHEELS, Int32ub),
    'gear_pos': Array(FG_MAX_WHEELS, Float32b),
    'gear_steer': Array(FG_MAX_WHEELS, Float32b),
    'gear_compression': Array(FG_MAX_WHEELS, Float32b),
    'cur_time_s': Int32ub,  # current unix time
    'warp_s': Int32sb,  # offset in seconds to unix time
    'visibility_m': Float32b,
    'elevator': Float32b,
    'elevator_trim_tab': Float32b,
    'left_flap': Float32b,
    'right_flap': Float32b,
    'left_aileron': Float32b,
    'right_aileron': Float32b,
    'rudder': Float32b,
    'nose_wheel': Float32b,
    'speedbrake': Float32b,
    'spoilers': Float32b,
}



"""
Main FlightGear interface module
"""
import copy
import socket
import sys
import re
import multiprocessing as mp
from typing import Callable, Optional, Tuple, Any, Dict, Union, ByteString

from construct import ConstError, Struct, Container

from .general_util import EventPipe, strip_end
from .fg_util import FGConnectionError, FGCommunicationError, fix_fg_radian_parsing

rx_callback_type = Callable[[Container, EventPipe], Optional[Container]]
"""
RX callback function type, signature should be:

.. code-block:: python

    def rx_cb(fdm_data: Construct.Container, event_pipe: EventPipe) -> Optional[Construct.Container]:
"""


class FGConnection:
    """
    Base class for FlightGear connections
    sphinx-no-autodoc
    """
    fg_net_struct: Optional[Struct] = None

    def __init__(self, rx_timeout_s: float = 2.0):
        self.event_pipe = EventPipe(duplex=True)

        self.fg_rx_sock: Optional[socket.socket] = None
        self.fg_rx_cb: Optional[rx_callback_type] = None

        self.fg_tx_sock: Optional[socket.socket] = None
        self.fg_tx_addr: Optional[Tuple[str, int]] = None

        self.rx_proc: Optional[mp.Process] = None
        self.rx_timeout_s = rx_timeout_s

    def connect_rx(self, fg_host: str, fg_port: int, rx_cb: rx_callback_type) -> EventPipe:
        """
        Connect to a UDP output of FlightGear

        :param fg_host: IP address of FG (usually localhost)
        :param fg_port: Port of the output socket (i.e. the ``5501`` from\
        ``--native-fdm=socket,out,30,,5501,udp``)
        :param rx_cb: Callback function, called whenever we receive data from FG.\
        Function signature should follow :attr:`rx_callback_type`
        :return: ``EventPipe`` so that data can be passed from the parent process\
        to the callback process
        """
        # TODO: Support TCP server so that we only need 1 port
        self.fg_rx_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        fg_rx_addr = (fg_host, fg_port)
        try:
            self.fg_rx_sock.bind(fg_rx_addr)
        except Exception as e:
            raise FGConnectionError(f'Could not bind to {fg_rx_addr}: {e}')
        self.fg_rx_cb = rx_cb

        return self.event_pipe

    def connect_tx(self, fg_host: str, fg_port: int):
        """
        Connect to a UDP input of FlightGear

        :param fg_host: IP address of FG (usually localhost)
        :param fg_port: Port of the input socket (i.e. the ``5502`` from\
        ``--native-fdm=socket,in,,,5502,udp``)
        """
        self.fg_tx_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.fg_tx_addr = (fg_host, fg_port)

    def _fg_packet_roundtrip(self):
        # Receive up to 1KB of data from FG
        # blocking is fine here since we're in a separate process
        try:
            rx_msg, _ = self.fg_rx_sock.recvfrom(1024)
        except socket.timeout as e:
            raise FGConnectionError(f'Timeout waiting for data, waited {self.rx_timeout_s} seconds') from e
        try:
            s: Container = self.fg_net_struct.parse(rx_msg)
        except ConstError as e:
            raise FGCommunicationError(f'Could not decode FG stream. Is this the right FDM version?\n{e}') from e

        if isinstance(self, FDMConnection):
            # Fix FG's radian parsing error :(
            s = fix_fg_radian_parsing(s)

        # Call user method
        s = self.fg_rx_cb(s, self.event_pipe)
        sys.stdout.flush()  # flush so that `print()` works

        # Send data back to FG
        if self.fg_tx_sock is not None and s is not None:
            tx_msg = self.fg_net_struct.build(dict(**s))
            self.fg_tx_sock.sendto(tx_msg, self.fg_tx_addr)

    def _rx_process(self):
        if self.fg_tx_sock is None:
            print(f'Warning: TX not connected, not sending updates to FG for RX {self.fg_rx_sock.getsockname()}')

        self.fg_rx_sock.settimeout(self.rx_timeout_s)
        self.event_pipe.child_send((True,))  # Signal to parent that child is running
        while True:
            self._fg_packet_roundtrip()

    def start(self):
        """
        Start the RX/TX loop with FlightGear
        """
        self.rx_proc = mp.Process(target=self._rx_process)
        self.rx_proc.start()
        _ = self.event_pipe.parent_recv()  # Wait for child to actually run

    def stop(self):
        """
        Stop the RX/TX loop
        """
        self.rx_proc.terminate()


class FDMConnection(FGConnection):
    """
    FlightGear Flight Dynamics Model Connection

    :param fdm_version: Net FDM version (24 or 25)
    """

    def __init__(self, fdm_version: int):
        super().__init__()
        # TODO: Support auto-version check
        if fdm_version == 24:
            from .fdm_v24 import fdm_struct
        elif fdm_version == 25:
            from .fdm_v25 import fdm_struct
        else:
            raise NotImplementedError(f'FDM version {fdm_version} not supported yet')
        # Create Struct from Dict
        self.fg_net_struct = Struct(*[k / v for k, v in fdm_struct.items()])


class CtrlsConnection(FGConnection):
    """
    FlightGear Controls Connection

    :param ctrls_version: Net Ctrls version (27)
    """

    def __init__(self, ctrls_version: int):
        super().__init__()
        # TODO: Support auto-version check
        if ctrls_version == 27:
            from .ctrls_v27 import ctrls_struct
        else:
            raise NotImplementedError(f'Controls version {ctrls_version} not supported yet')
        # Create Struct from Dict
        self.fg_net_struct = Struct(*[k / v for k, v in ctrls_struct.items()])


class PropsConnection:
    """
    FlightGear Telnet Interface Connection (also known as the property interface).
    See https://wiki.flightgear.org/Telnet_usage for general details.

    :param host: IP address of FG (usually localhost)
    :param tcp_port: Port of the telnet socket (i.e. the ``5500`` from\
        ``--telnet=socket,bi,60,localhost,5500,tcp``)
    :param rx_timeout_s: Optional timeout value in seconds when recieving data
    """

    def __init__(self, host: str, tcp_port: int, rx_timeout_s: float = 2.0):
        self.host = host
        self.port = tcp_port
        # SOCK_STREAM == TCP
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.rx_timeout_s = rx_timeout_s

    def connect(self):
        """
        Connect to the FlightGear telnet server
        """
        telnet_addr = (self.host, self.port)
        try:
            self.sock.connect(telnet_addr)
        except Exception as e:
            raise FGConnectionError(f'Could not connect to FlightGear telnet server {telnet_addr}: {e}') from e
        # force move to the root directory (maybe someone was connected before we were)
        _ = self._send_cmd_get_resp('cd /')

    @staticmethod
    def _telnet_str(in_str: str) -> ByteString:
        return f'{in_str}\r\n'.encode()

    def _send_cmd_get_resp(self, cmd_str: str, buflen: int = 512) -> str:
        self.sock.sendall(self._telnet_str(cmd_str))

        self.sock.settimeout(self.rx_timeout_s)
        # FG telnet always ends with a prompt (`cwd`> ), and since we always
        # operate relative to the root directory, it should always be the same prompt
        ending_bytes = b'/> '
        resp_bytes = b''
        while not resp_bytes.endswith(ending_bytes):
            # Loop until FG sends us all the data
            try:
                resp_bytes += self.sock.recv(buflen)
            except socket.timeout as e:
                raise FGConnectionError(f'Timeout waiting for data, waited {self.rx_timeout_s} seconds') from e
        resp_bytes = strip_end(resp_bytes, b'\r\n' + ending_bytes)  # trim the prompt

        resp_str = resp_bytes.decode()
        if resp_str.startswith('-ERR'):
            raise FGCommunicationError(f'Bad telnet command "{cmd_str}". Response: "{resp_str}"')
        return resp_str

    @staticmethod
    def _extract_fg_prop(resp_str: str) -> Tuple[str, Any]:
        try:
            match = re.search(r"^(.+)\s=\s+'(.*)'\s+\((.+)\)$", resp_str, flags=re.DOTALL)
        except Exception as e:
            raise FGCommunicationError(f'Could not parse FG telnet response for msg "{resp_str}": {e}') from e
        if match is None:
            raise FGCommunicationError(f'Could not parse FG telnet response for msg "{resp_str}"')
        key_str = match.group(1)
        value_str = match.group(2)
        type_str = match.group(3)
        convert_fn = {
            'bool': bool,
            'int': int,
            'string': str,
            'double': float,
        }.get(type_str, lambda x: x)
        try:
            value = convert_fn(value_str)
        except ValueError as e:
            raise FGCommunicationError(f'Could not auto-convert "{resp_str}": {e}') from e
        return key_str, value

    def get_prop(self, prop_str: str) -> Any:
        """
        Get a property from FlightGear.

        :param prop_str: Location of the property, should always be relative to\
            the root (``/``)
        :return: The value of the property. If FG tells us what the type is we \
            will pre-convert it (i.e. make an int from a string)
        """
        if not prop_str.startswith('/'):
            raise ValueError(f'Property must be absolute (start with /): {prop_str}')
        resp_str = self._send_cmd_get_resp(f'get {prop_str}')
        _, value = self._extract_fg_prop(resp_str)
        return value

    def set_prop(self, prop_str: str, value: Any):
        """
        Set a property in FlightGear.

        :param prop_str: Location of the property, should always be relative to\
            the root (``/``)
        :param value: Value to set the property to. Must be convertible to ``str``
        """
        if not prop_str.startswith('/'):
            raise ValueError(f'Property must be absolute (start with /): {prop_str}')
        _ = self._send_cmd_get_resp(f'set {prop_str} {str(value)}')
        # We don't care about the response

    def list_props(self, path: str = '/', recurse_limit: Optional[int] = 0) -> Dict[str, Union[list, Dict]]:
        """
        List properties in the FlightGear property tree.

        :param path: Directory to list from, should always be relative to
            the root (``/``)
        :param recurse_limit: How many times to recurse into subdirectories.
            1 (default) is no recursion, 2 is 1 level deep, etc. Passing in
            ``None`` disables the recursion limit. Be warned that enabling any kind of recursion will take a long time!
        :return: Dictionary with keys:

            * ``directories``: List of directories, absolute path
            * ``properties``: Dictionary with property name as the key (absolute path), value as their value.

        Example for ``list_props('/position', recurse_limit=0)``:

        .. code-block:: python

            {
                'directories': [
                    '/position/model'
                ],
                'properties': {
                    '/position/altitude-agl-ft': 3.148566963,
                    '/position/altitude-agl-m': 0.9596832103,
                    '/position/altitude-ft': 3491.986254,
                    '/position/ground-elev-ft': 3488.469757,
                    '/position/ground-elev-m': 1063.285582,
                    '/position/latitude-deg': 0.104476136,
                    '/position/latitude-string': '0*06\\'16.1"N',
                    '/position/longitude-deg': 100.023135,
                    '/position/longitude-string': '100*01\\'23.3"E',
                    '/position/sea-level-radius-ft': 20925646.09
                }
            }
        """
        if not path.startswith('/'):
            raise ValueError(f'Path must be absolute (start with /): {path}')
        path = path.rstrip('/')  # Strip trailing slash to keep things consistent

        resp_list = self._send_cmd_get_resp(f'ls {path}').split('\r\n')
        # extract of directories, absolute path
        dir_list = [f'{path}/{s.rstrip("/")}' for s in resp_list if s.endswith('/')]

        prop_dict = {}
        # recursion support
        if recurse_limit is None or recurse_limit > 0:
            dir_list_cwd = copy.deepcopy(dir_list)
            for dir_str in dir_list_cwd:
                # Handle None as a recursion limit
                if recurse_limit is None:
                    new_recurse_limit = None
                else:
                    new_recurse_limit = recurse_limit - 1
                # recursion!
                dir_dict = self.list_props(dir_str, recurse_limit=new_recurse_limit)
                # add the recursed properties
                prop_dict = {**prop_dict, **dir_dict['properties']}
                dir_list.remove(dir_str)  # remove the non-recursed directory
                # add the recursed directories
                dir_list += dir_dict['directories']

        # extract all the values
        for s in resp_list:
            if '=' in s:
                key, val = self._extract_fg_prop(s)
                # prepend the key with the working directory, keep naming consistent
                prop_dict[f'{path}/{key}'] = val

        # Returned paths are absolute
        rtn_dict = {
            # sort the list because we're nice
            'directories': dir_list,
            'properties': prop_dict,
        }
        return rtn_dict



"""
FlightGear Controls Network interface, version 27

See https://github.com/FlightGear/flightgear/blob/619226e9d069d2a3e8ebf8658fb5441ca8a2c233/src/Network/net_ctrls.hxx
"""

from construct import Array, Enum, Const, Padding, Int32ub, Float64b, BitStruct, Flag, BitsInteger, Mapping

RESERVED_SPACE = 25  #: Constant value from define

FG_MAX_ENGINES = 4  #: Constant value from enum
# FG_MAX_WHEELS = 16  #: Constant value from enum. TODO: Unused?
FG_MAX_TANKS = 8  #: Constant value from enum

# Big endian
#: Ctrls v27 structure
ctrls_struct = {
    'version': Const(27, Int32ub),
    '_padding0': Padding(4),  # TODO: Not documented, probably due to struct packing
    'aileron': Float64b,  # -1 ... 1
    'elevator': Float64b,  # -1 ... 1
    'rudder': Float64b,  # -1 ... 1
    'aileron_trim': Float64b,  # -1 ... 1
    'elevator_trim': Float64b,  # -1 ... 1
    'rudder_trim': Float64b,  # -1 ... 1
    'flaps': Float64b,  # 0 ... 1
    'spoilers': Float64b,
    'speedbrake': Float64b,
    'flaps_power': Enum(Int32ub, unavailable=0, available=1),
    'flap_motor_ok': Int32ub,
    'num_engines': Int32ub,
    'master_bat': Array(FG_MAX_ENGINES, Enum(Int32ub, off=0, on=1)),
    'master_alt': Array(FG_MAX_ENGINES, Enum(Int32ub, off=0, on=1)),
    'magnetos': Array(FG_MAX_ENGINES, Int32ub),
    'starter_power': Array(FG_MAX_ENGINES,
                           Enum(Int32ub, off=0, on=1)
                           ),
    # throttle needs to be moved forward 1 double?
    '_padding3': Padding(4),  # TODO: Not documented, probably due to struct packing
    'throttle': Array(FG_MAX_ENGINES, Float64b),  # 0 ... 1
    'mixture': Array(FG_MAX_ENGINES, Float64b),  # 0 ... 1
    'condition': Array(FG_MAX_ENGINES, Float64b),  # 0 ... 1
    'fuel_pump_power': Array(FG_MAX_ENGINES,
                             Enum(Int32ub, off=0, on=1)
                             ),
    'prop_advance': Array(FG_MAX_ENGINES, Float64b),  # 0 ... 1
    'feed_tank_to': Array(4, Int32ub),
    'reverse': Array(4, Int32ub),
    'engine_ok': Array(FG_MAX_ENGINES, Int32ub),
    'mag_left_ok': Array(FG_MAX_ENGINES, Int32ub),
    'mag_right_ok': Array(FG_MAX_ENGINES, Int32ub),
    'spark_plugs_ok': Array(FG_MAX_ENGINES, Enum(Int32ub, fouled=0, ok=1)),
    'oil_press_status': Array(FG_MAX_ENGINES,
                              Enum(Int32ub, normal=0, low=1, full_fail=2)
                              ),
    'fuel_pump_ok': Array(FG_MAX_ENGINES, Int32ub),
    'num_tanks': Int32ub,
    'fuel_selector': Array(FG_MAX_TANKS, Enum(Int32ub, off=0, on=1)),
    'xfer_pump': Array(5, Int32ub),  # specifies transfer from array value tank to tank specified by int value
    'cross_feed': Enum(Int32ub, off=0, on=1),
    '_padding4': Padding(4),  # TODO: Not documented, probably due to struct packing
    'brake_left': Float64b,
    'brake_right': Float64b,
    'copilot_brake_left': Float64b,
    'copilot_brake_right': Float64b,
    'brake_parking': Float64b,
    'gear_handle': Enum(Int32ub, up=0, down=1),
    'master_avionics': Enum(Int32ub, off=0, on=1),
    'comm_1_MHz': Float64b,  # TODO: Frequencies are all screwed up. Garbage data
    'comm_2_MHz': Float64b,
    'nav_1_MHz': Float64b,
    'nav_2_MHz': Float64b,
    'wind_speed_kt': Float64b,
    'wind_dir_deg': Float64b,
    'turbulence_norm': Float64b,
    'temp_c': Float64b,
    'press_inhg': Float64b,
    'hground_m': Float64b,  # ground elevation
    'magvar_deg': Float64b,  # local magnetic variation
    'icing': Int32ub,
    'speedup': Int32ub,  # integer speedup multiplier
    'freeze': BitStruct(  # Default is big-endian
        other=BitsInteger((Int32ub.sizeof() * 8) - 3),  # Rest of uint32 minus the 3 flags
        fuel=Flag,
        position=Flag,
        master=Flag,
    ),
    '_reserved': Padding(Int32ub.length * RESERVED_SPACE),
}



"""
FlightGear-specific utility functionality
"""
import math

from construct import Container


class FGConnectionError(Exception):
    """
    Error type generated when there is an error connecting to FlightGear
    """
    pass


class FGCommunicationError(Exception):
    """
    Error type generated when there is an error talking to FlightGear
    """
    pass


def offset_fg_radian(in_rad: float) -> float:
    """
    Even when echoing back literally what FG sends over the Net FDM connection,
    (i.e. UDP bytes in -> UDP bytes out) the latitude/longitude shown in FG
    appear to decrease. After plotting the offsets at a couple different lat/lons
    it appears to be a linear relationship and identical in anything that is
    represented in radians. The coefficient was chosen through trial-and-error.
    sphinx-no-autodoc

    :param in_rad: Input property, in radians
    :return: Offset that needs to be applied to the input, in radians
    """
    coeff = 1.09349403e-9
    return math.degrees(in_rad) * coeff


def fix_fg_radian_parsing(s: Container) -> Container:
    """
    Helper for all the radian values in the FDM
    sphinx-no-autodoc
    """
    s.lon_rad += offset_fg_radian(s.lon_rad)
    s.lat_rad += offset_fg_radian(s.lat_rad)
    s.phi_rad += offset_fg_radian(s.phi_rad)
    s.theta_rad += offset_fg_radian(s.theta_rad)
    s.psi_rad += offset_fg_radian(s.psi_rad)
    s.alpha_rad += offset_fg_radian(s.alpha_rad)
    s.beta_rad += offset_fg_radian(s.beta_rad)
    s.phidot_rad_per_s += offset_fg_radian(s.phidot_rad_per_s)
    s.thetadot_rad_per_s += offset_fg_radian(s.thetadot_rad_per_s)
    s.psidot_rad_per_s += offset_fg_radian(s.psidot_rad_per_s)
    return s



"""
non-FlightGear-specific utility functionality
"""
import multiprocessing as mp
from typing import Any, Union, ByteString


class EventPipe:
    """
    Helper abstraction for passing data from a parent process to a child
    process.

    :param duplex: Allow internal pipe to also pass data from child to \
    parent. Recommended to leave at default
    """

    def __init__(self, duplex=True):
        self.event = mp.Event()
        # If not duplex, can only transfer data from parent to child
        self.child_pipe, self.parent_pipe = mp.Pipe(duplex=duplex)

        # function aliases
        self.set = self.event.set
        self.is_set = self.event.is_set
        self.clear = self.event.clear
        self.child_send = self.child_pipe.send
        self.child_poll = self.child_pipe.poll
        self.parent_recv = self.parent_pipe.recv
        self.parent_poll = self.parent_pipe.poll

    def parent_send(self, *args, **kwargs):
        """
        Send data from the parent process to the child process, then set\
        our event flag

        :param args: Passed to :meth:`multiprocessing.connection.Connection.send()`
        :param kwargs: Passed to :meth:`multiprocessing.connection.Connection.send()`
        """
        if not self.is_set():
            # Only send when data has been received
            self.parent_pipe.send(*args, **kwargs)
            self.set()

    def child_recv(self, *args, **kwargs) -> Any:
        """
        Receive data from the parent process to the child process, then clear\
        our event flag

        :param args: Passed to :meth:`multiprocessing.connection.Connection.recv()`
        :param kwargs: Passed to :meth:`multiprocessing.connection.Connection.recv()`
        """
        msg = self.child_pipe.recv(*args, **kwargs)
        self.clear()
        return msg


def strip_end(text: Union[str, ByteString], suffix: Union[str, ByteString]) -> Union[str, ByteString]:
    """
    This could be removed if we want to move lowest supported version to 3.9 (.removesuffix())
    sphinx-no-autodoc

    :param text: text to strip the end from
    :param suffix: string to remove from the end
    :return: text with suffix removed
    """
    if suffix and text.endswith(suffix):
        return text[:-len(suffix)]
    return text



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flightgear_python']

package_data = \
{'': ['*']}

install_requires = \
['construct>=2.10,<3.0']

setup_kwargs = {
    'name': 'flightgear-python',
    'version': '1.2.0',
    'description': 'Interface for FlightGear network connections',
    'long_description': '# FlightGear Python Interface\n[![Documentation Status](https://readthedocs.org/projects/flightgear-python/badge/?version=latest)](https://flightgear-python.readthedocs.io/en/latest/?badge=latest)\n[![CircleCI](https://circleci.com/gh/julianneswinoga/flightgear-python.svg?style=shield)](https://circleci.com/gh/julianneswinoga/flightgear-python)\n[![Coverage Status](https://coveralls.io/repos/github/julianneswinoga/flightgear-python/badge.svg?branch=master)](https://coveralls.io/github/julianneswinoga/flightgear-python?branch=master)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/flightgear_python)](https://pypi.org/project/flightgear-python/)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/flightgear-python)\n\n`flightgear-python` is an interface package to the [FlightGear flight simulation software](https://www.flightgear.org/) aimed at simplicity.\n\nInstall: `pip3 install flightgear-python`\n\nDon\'t know where to begin? Check out the [quick-start](https://flightgear-python.readthedocs.io/en/latest/quickstart.html) documentation.\n\nFDM example, from `examples/simple_fdm.py`\n```python\n"""\nSimple Flight Dynamics Model (FDM) example that makes the altitude increase and the plane roll in the air.\n"""\nimport time\nfrom flightgear_python.fg_if import FDMConnection\n\ndef fdm_callback(fdm_data, event_pipe):\n    if event_pipe.child_poll():\n        phi_rad_child, = event_pipe.child_recv()  # unpack tuple\n        # set only the data that we need to\n        fdm_data[\'phi_rad\'] = phi_rad_child  # we can force our own values\n    fdm_data.alt_m = fdm_data.alt_m + 0.5  # or just make a relative change\n    return fdm_data  # return the whole structure\n\n"""\nStart FlightGear with `--native-fdm=socket,out,30,,5501,udp --native-fdm=socket,in,30,,5502,udp`\n(you probably also want `--fdm=null` and `--max-fps=30` to stop the simulation fighting with\nthese external commands)\n"""\nfdm_conn = FDMConnection(fdm_version=24)  # May need to change version from 24\nfdm_event_pipe = fdm_conn.connect_rx(\'localhost\', 5501, fdm_callback)\nfdm_conn.connect_tx(\'localhost\', 5502)\nfdm_conn.start()  # Start the FDM RX/TX loop\n\nphi_rad_parent = 0.0\nwhile True:\n    phi_rad_parent += 0.1\n    # could also do `fdm_conn.event_pipe.parent_send` so you just need to pass around `fdm_conn`\n    fdm_event_pipe.parent_send((phi_rad_parent,))  # send tuple\n    time.sleep(0.5)\n```\n\nSupported interfaces:\n- [x] [Native Protocol](https://wiki.flightgear.org/Property_Tree/Sockets) (currently only UDP)\n  - [x] Flight Dynamics Model ([`net_fdm.hxx`](https://github.com/FlightGear/flightgear/blob/next/src/Network/net_fdm.hxx))\n  - [x] Controls ([`net_ctrls.hxx`](https://github.com/FlightGear/flightgear/blob/next/src/Network/net_ctrls.hxx))\n  - [ ] GUI ([`net_gui.hxx`](https://github.com/FlightGear/flightgear/blob/next/src/Network/net_gui.hxx))\n- [ ] [Generic Protocol](https://wiki.flightgear.org/Generic_protocol)\n- [x] [Telnet](https://wiki.flightgear.org/Telnet_usage)\n- [ ] [HTTP](https://wiki.flightgear.org/Property_Tree_Servers)\n',
    'author': 'Julianne Swinoga',
    'author_email': 'julianneswinoga@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
